@@ -1,19 +1,21 @@
-import abc
+from __future__ import annotations
+
 from dataclasses import dataclass
 from typing import Any, Dict, List
 
-from qoala.lang.iqoala import IqoalaProgram
-from qoala.runtime.schedule import ProgramTaskList
-
-
-class ProgramContext(abc.ABC):
-    pass
+from qoala.lang.ehi import UnitModule
+from qoala.lang.program import QoalaProgram
+from qoala.runtime.schedule import BlockTask
 
 
 @dataclass
 class ProgramInput:
     values: Dict[str, Any]
 
+    @classmethod
+    def empty(cls) -> ProgramInput:
+        return ProgramInput({})
+
 
 @dataclass
 class ProgramResult:
@@ -24,22 +26,22 @@ class ProgramResult:
 class BatchInfo:
     """Description of a batch of program instances that should be executed."""
 
-    program: IqoalaProgram
+    program: QoalaProgram
+    unit_module: UnitModule
     inputs: List[ProgramInput]  # dict of inputs for each iteration
     num_iterations: int
     deadline: float
-    tasks: ProgramTaskList
-    num_qubits: int  # TODO: replace this by unit module
 
 
 @dataclass
 class ProgramInstance:
-    """A running program"""
+    """A program instantiated with Program Inputs and a Unit Module"""
 
     pid: int
-    program: IqoalaProgram
+    program: QoalaProgram
     inputs: ProgramInput
-    tasks: ProgramTaskList
+    unit_module: UnitModule
+    block_tasks: List[BlockTask]
 
 
 @dataclass
