@@ -1,70 +1,60 @@
 from __future__ import annotations
 
 from dataclasses import dataclass
-from typing import Any, Dict, Generator, List, Optional, Tuple, Type
+from typing import Any, Dict, Generator, List, Optional, Type
 
 import netsquid as ns
-from netqasm.lang.instr.core import CreateEPRInstruction, RecvEPRInstruction
 from netqasm.lang.parsing import parse_text_subroutine
-from netqasm.sdk.build_epr import SER_RESPONSE_KEEP_LEN
 from netsquid.components import QuantumProcessor
-from netsquid.nodes import Node
 from netsquid.qubits import ketstates
-from netsquid.qubits.ketstates import BellIndex
-from netsquid_magic.link_layer import (
-    MagicLinkLayerProtocolWithSignaling,
-    SingleClickTranslationUnit,
-)
-from netsquid_magic.magic_distributor import PerfectStateMagicDistributor
-from qlink_interface import ReqCreateBase, ResCreateAndKeep
-from qlink_interface.interface import ResCreate
 
 from pydynaa import EventExpression
-from qoala.lang.iqoala import (
+from qoala.lang.ehi import EhiNetworkInfo, UnitModule
+from qoala.lang.hostlang import (
     AssignCValueOp,
+    BasicBlock,
+    BasicBlockType,
     ClassicalIqoalaOp,
-    IqoalaParser,
-    IqoalaProgram,
-    IqoalaRequest,
-    IqoalaSubroutine,
-    IQoalaSubroutineParser,
     IqoalaVector,
-    ProgramMeta,
     ReceiveCMsgOp,
     RunSubroutineOp,
     SendCMsgOp,
 )
+from qoala.lang.parse import LocalRoutineParser, QoalaParser
+from qoala.lang.program import LocalRoutine, ProgramMeta, QoalaProgram
+from qoala.lang.request import (
+    CallbackType,
+    EprRole,
+    EprType,
+    QoalaRequest,
+    RequestRoutine,
+    RequestVirtIdMapping,
+)
+from qoala.lang.routine import RoutineMetadata
 from qoala.runtime.config import GenericQDeviceConfig
-from qoala.runtime.environment import (
-    GlobalEnvironment,
-    GlobalNodeInfo,
-    LocalEnvironment,
+from qoala.runtime.environment import LocalEnvironment, NetworkInfo
+from qoala.runtime.lhi import LhiLatencies, LhiLinkInfo, LhiTopology, LhiTopologyBuilder
+from qoala.runtime.lhi_to_ehi import (
+    GenericToVanillaInterface,
+    LhiConverter,
+    NativeToFlavourInterface,
 )
+from qoala.runtime.memory import ProgramMemory
+from qoala.runtime.message import Message, RrCallTuple
 from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
-from qoala.runtime.schedule import ProgramTaskList
 from qoala.sim.build import build_generic_qprocessor
-from qoala.sim.csocket import ClassicalSocket
-from qoala.sim.egp import EgpProtocol
-from qoala.sim.hostinterface import HostInterface
-from qoala.sim.memmgr import AllocError, MemoryManager
-from qoala.sim.memory import ProgramMemory, SharedMemory, Topology, UnitModule
-from qoala.sim.message import Message
-from qoala.sim.netstackinterface import NetstackInterface
-from qoala.sim.process import IqoalaProcess
+from qoala.sim.entdist.entdist import EntDist
+from qoala.sim.entdist.entdistcomp import EntDistComponent
+from qoala.sim.events import MSG_REQUEST_DELIVERED
+from qoala.sim.host.csocket import ClassicalSocket
+from qoala.sim.host.hostinterface import HostInterface
+from qoala.sim.memmgr import MemoryManager
+from qoala.sim.netstack import NetstackInterface
+from qoala.sim.process import QoalaProcess
 from qoala.sim.procnode import ProcNode
-from qoala.sim.qdevice import (
-    PhysicalQuantumMemory,
-    QDevice,
-    QDeviceCommand,
-    QDeviceType,
-)
-from qoala.sim.qnosinterface import QnosInterface
-from qoala.sim.requests import (
-    EprCreateType,
-    NetstackCreateRequest,
-    NetstackReceiveRequest,
-)
-from qoala.util.tests import has_multi_state, netsquid_run
+from qoala.sim.qdevice import QDevice, QDeviceCommand
+from qoala.util.math import has_multi_state
+from qoala.util.tests import netsquid_run
 
 MOCK_MESSAGE = Message(content=42)
 MOCK_QNOS_RET_REG = "R0"
@@ -93,34 +83,10 @@ class MockNetstackInterface(NetstackInterface):
         local_env: LocalEnvironment,
         qdevice: QDevice,
         memmgr: MemoryManager,
-        mock_result: ResCreate,
     ) -> None:
         self._qdevice = qdevice
         self._local_env = local_env
         self._memmgr = memmgr
-        self._mock_result = mock_result
-
-        self._requests_put: Dict[int, List[ReqCreateBase]] = {}
-        self._awaited_result_ck: List[int] = []  # list of remote ids
-        self._awaited_mem_free_sig_count: int = 0
-
-    def put_request(self, remote_id: int, request: ReqCreateBase) -> None:
-        if remote_id not in self._requests_put:
-            self._requests_put[remote_id] = []
-        self._requests_put[remote_id].append(request)
-
-    def await_result_create_keep(
-        self, remote_id: int
-    ) -> Generator[EventExpression, None, ResCreateAndKeep]:
-        self._awaited_result_ck.append(remote_id)
-        return self._mock_result
-        yield  # to make this behave as a generator
-
-    def await_memory_freed_signal(
-        self, pid: int, virt_id: int
-    ) -> Generator[EventExpression, None, None]:
-        raise AllocError
-        yield  # to make this behave as a generator
 
     def send_qnos_msg(self, msg: Message) -> None:
         return None
@@ -132,22 +98,24 @@ class MockNetstackInterface(NetstackInterface):
         return None
         yield  # to make this behave as a generator
 
-    def reset(self) -> None:
-        self._requests_put = {}
-        self._awaited_result_ck = []
-        self._awaited_mem_free_sig_count = 0
+    def send_entdist_msg(self, msg: Message) -> None:
+        return None
+
+    def receive_entdist_msg(self) -> Generator[EventExpression, None, Message]:
+        return Message(MSG_REQUEST_DELIVERED)
+        yield  # to make this behave as a generator
+
+    @property
+    def node_id(self) -> int:
+        return 0
 
 
 class MockQDevice(QDevice):
-    def __init__(self, topology: Topology) -> None:
-        self._memory = PhysicalQuantumMemory(topology.comm_ids, topology.mem_ids)
+    def __init__(self, topology: LhiTopology) -> None:
+        self._topology = topology
 
         self._executed_commands: List[QDeviceCommand] = []
 
-    @property
-    def typ(self) -> QDeviceType:
-        return QDeviceType.GENERIC
-
     def set_mem_pos_in_use(self, id: int, in_use: bool) -> None:
         pass
 
@@ -162,83 +130,11 @@ class MockQDevice(QDevice):
         self._executed_commands = []
 
 
-@dataclass
-class MockNetstackResultInfo:
-    pid: int
-    array_id: int
-    start_idx: int
-    end_idx: int
-
-
-class MockQnosInterface(QnosInterface):
-    def __init__(
-        self,
-        qdevice: QDevice,
-        netstack_result_info: Optional[MockNetstackResultInfo] = None,
-    ) -> None:
-        self.send_events: List[InterfaceEvent] = []
-        self.recv_events: List[InterfaceEvent] = []
-        self.flush_events: List[FlushEvent] = []
-        self.signal_events: List[SignalEvent] = []
-
-        self._qdevice = qdevice
-        self._memmgr = MemoryManager("alice", self._qdevice)
-
-        self.netstack_result_info: Optional[
-            MockNetstackResultInfo
-        ] = netstack_result_info
-
-    def send_peer_msg(self, peer: str, msg: Message) -> None:
-        self.send_events.append(InterfaceEvent(peer, msg))
-
-    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
-        self.recv_events.append(InterfaceEvent(peer, MOCK_MESSAGE))
-        return MOCK_MESSAGE
-        yield  # to make it behave as a generator
-
-    def send_host_msg(self, msg: Message) -> None:
-        self.send_events.append(InterfaceEvent("host", msg))
-
-    def receive_host_msg(self) -> Generator[EventExpression, None, Message]:
-        self.recv_events.append(InterfaceEvent("host", MOCK_MESSAGE))
-        return MOCK_MESSAGE
-        yield  # to make it behave as a generator
-
-    def send_netstack_msg(self, msg: Message) -> None:
-        self.send_events.append(InterfaceEvent("netstack", msg))
-
-    def receive_netstack_msg(self) -> Generator[EventExpression, None, Message]:
-        self.recv_events.append(InterfaceEvent("netstack", MOCK_MESSAGE))
-        if self.netstack_result_info is not None:
-            mem = self.memmgr._processes[
-                self.netstack_result_info.pid
-            ].prog_memory.shared_mem
-            array_id = self.netstack_result_info.array_id
-            start_idx = self.netstack_result_info.start_idx
-            end_idx = self.netstack_result_info.end_idx
-            for i in range(start_idx, end_idx):
-                mem.set_array_value(array_id, i, 42)
-        return MOCK_MESSAGE
-        yield  # to make it behave as a generator
-
-    def flush_netstack_msgs(self) -> None:
-        self.flush_events.append(FlushEvent())
-
-    def signal_memory_freed(self) -> None:
-        self.signal_events.append(SignalEvent())
-
-    @property
-    def name(self) -> str:
-        return "mock"
-
-
 class MockHostInterface(HostInterface):
-    def __init__(self, shared_mem: Optional[SharedMemory] = None) -> None:
+    def __init__(self) -> None:
         self.send_events: List[InterfaceEvent] = []
         self.recv_events: List[InterfaceEvent] = []
 
-        self.shared_mem = shared_mem
-
     def send_peer_msg(self, peer: str, msg: Message) -> None:
         self.send_events.append(InterfaceEvent(peer, msg))
 
@@ -252,8 +148,6 @@ class MockHostInterface(HostInterface):
 
     def receive_qnos_msg(self) -> Generator[EventExpression, None, Message]:
         self.recv_events.append(InterfaceEvent("qnos", MOCK_MESSAGE))
-        if self.shared_mem is not None:
-            self.shared_mem.set_reg_value(MOCK_QNOS_RET_REG, MOCK_QNOS_RET_VALUE)
         return MOCK_MESSAGE
         yield  # to make it behave as a generator
 
@@ -264,31 +158,36 @@ class MockHostInterface(HostInterface):
 
 def create_program(
     instrs: Optional[List[ClassicalIqoalaOp]] = None,
-    subroutines: Optional[Dict[str, IqoalaSubroutine]] = None,
-    requests: Optional[Dict[str, IqoalaRequest]] = None,
+    subroutines: Optional[Dict[str, LocalRoutine]] = None,
+    req_routines: Optional[Dict[str, RequestRoutine]] = None,
     meta: Optional[ProgramMeta] = None,
-) -> IqoalaProgram:
+) -> QoalaProgram:
     if instrs is None:
         instrs = []
     if subroutines is None:
         subroutines = {}
 
-    if requests is None:
-        requests = {}
+    if req_routines is None:
+        req_routines = {}
     if meta is None:
         meta = ProgramMeta.empty("prog")
-    return IqoalaProgram(
-        instructions=instrs, subroutines=subroutines, meta=meta, requests=requests
+    # TODO: split into proper blocks
+    block = BasicBlock("b0", BasicBlockType.CL, instrs)
+    return QoalaProgram(
+        blocks=[block],
+        local_routines=subroutines,
+        meta=meta,
+        request_routines=req_routines,
     )
 
 
 def create_process(
     pid: int,
-    program: IqoalaProgram,
+    program: QoalaProgram,
     unit_module: UnitModule,
     host_interface: HostInterface,
     inputs: Optional[Dict[str, Any]] = None,
-) -> IqoalaProcess:
+) -> QoalaProcess:
     if inputs is None:
         inputs = {}
     prog_input = ProgramInput(values=inputs)
@@ -296,11 +195,12 @@ def create_process(
         pid=pid,
         program=program,
         inputs=prog_input,
-        tasks=ProgramTaskList.empty(program),
+        unit_module=unit_module,
+        block_tasks=[],
     )
-    mem = ProgramMemory(pid=0, unit_module=unit_module)
+    mem = ProgramMemory(pid=0)
 
-    process = IqoalaProcess(
+    process = QoalaProcess(
         prog_instance=instance,
         prog_memory=mem,
         csockets={
@@ -308,8 +208,6 @@ def create_process(
             for (id, name) in program.meta.csockets.items()
         },
         epr_sockets=program.meta.epr_sockets,
-        subroutines=program.subroutines,
-        requests=program.requests,
         result=ProgramResult(values={}),
     )
     return process
@@ -320,20 +218,20 @@ def create_qprocessor(name: str, num_qubits: int) -> QuantumProcessor:
     return build_generic_qprocessor(name=f"{name}_processor", cfg=cfg)
 
 
-def create_global_env(
+def create_network_info(
     num_qubits: int, names: List[str] = ["alice", "bob", "charlie"]
-) -> GlobalEnvironment:
-
-    env = GlobalEnvironment()
-    for i, name in enumerate(names):
-        env.add_node(i, GlobalNodeInfo(name, i))
-    return env
+) -> NetworkInfo:
+    nodes = {i: name for i, name in enumerate(names)}
+    return NetworkInfo.with_nodes(nodes)
 
 
 def create_procnode(
     name: str,
-    env: GlobalEnvironment,
+    env: NetworkInfo,
     num_qubits: int,
+    topology: LhiTopology,
+    latencies: LhiLatencies,
+    ntf_interface: NativeToFlavourInterface,
     procnode_cls: Type[ProcNode] = ProcNode,
     asynchronous: bool = False,
 ) -> ProcNode:
@@ -342,8 +240,12 @@ def create_procnode(
     node_id = env.get_node_id(name)
     procnode = procnode_cls(
         name=name,
-        global_env=env,
+        network_info=env,
         qprocessor=alice_qprocessor,
+        qdevice_topology=topology,
+        latencies=latencies,
+        ntf_interface=ntf_interface,
+        network_ehi=EhiNetworkInfo({}),
         node_id=node_id,
         asynchronous=asynchronous,
     )
@@ -351,46 +253,66 @@ def create_procnode(
     return procnode
 
 
-def simple_subroutine(name: str, subrt_text: str) -> IqoalaSubroutine:
+def simple_subroutine(name: str, subrt_text: str) -> LocalRoutine:
     subrt = parse_text_subroutine(subrt_text)
-    return IqoalaSubroutine(name, subrt, return_map={})
+    return LocalRoutine(
+        name, subrt, return_vars=[], metadata=RoutineMetadata.use_none()
+    )
 
 
-def parse_iqoala_subroutines(subrt_text: str) -> IqoalaSubroutine:
-    return IQoalaSubroutineParser(subrt_text).parse()
+def parse_iqoala_subroutines(subrt_text: str) -> LocalRoutine:
+    return LocalRoutineParser(subrt_text).parse()
 
 
-def create_egp_protocols(node1: Node, node2: Node) -> Tuple[EgpProtocol, EgpProtocol]:
-    link_dist = PerfectStateMagicDistributor(nodes=[node1, node2], state_delay=1000.0)
-    link_prot = MagicLinkLayerProtocolWithSignaling(
-        nodes=[node1, node2],
-        magic_distributor=link_dist,
-        translation_unit=SingleClickTranslationUnit(),
+def generic_topology(num_qubits: int) -> LhiTopology:
+    # Instructions and durations are not needed for these tests.
+    return LhiTopologyBuilder.perfect_uniform(
+        num_qubits=num_qubits,
+        single_instructions=[],
+        single_duration=0,
+        two_instructions=[],
+        two_duration=0,
+    )
+
+
+def star_topology(num_qubits: int) -> LhiTopology:
+    # Instructions and durations are not needed for these tests.
+    return LhiTopologyBuilder.perfect_star(
+        num_qubits=num_qubits,
+        comm_instructions=[],
+        comm_duration=0,
+        mem_instructions=[],
+        mem_duration=0,
+        two_instructions=[],
+        two_duration=0,
     )
-    return EgpProtocol(node1, link_prot), EgpProtocol(node2, link_prot)
 
 
 def test_initialize():
     num_qubits = 3
-    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})
+    topology = generic_topology(num_qubits)
+    latencies = LhiLatencies.all_zero()
+    ntf = GenericToVanillaInterface()
 
-    global_env = create_global_env(num_qubits)
-    local_env = LocalEnvironment(global_env, global_env.get_node_id("alice"))
-    procnode = create_procnode("alice", global_env, num_qubits)
+    network_info = create_network_info(num_qubits)
+    local_env = LocalEnvironment(network_info, network_info.get_node_id("alice"))
+    procnode = create_procnode(
+        "alice", network_info, num_qubits, topology, latencies, ntf
+    )
     procnode.qdevice = MockQDevice(topology)
 
     procnode.host.interface = MockHostInterface()
 
-    mock_result = ResCreateAndKeep(bell_state=BellIndex.B01)
     procnode.netstack.interface = MockNetstackInterface(
-        local_env, procnode.qdevice, procnode.memmgr, mock_result
+        local_env, procnode.qdevice, procnode.memmgr
     )
 
     host_processor = procnode.host.processor
     qnos_processor = procnode.qnos.processor
     netstack_processor = procnode.netstack.processor
 
-    unit_module = UnitModule.default_generic(num_qubits)
+    ehi = LhiConverter.to_ehi(topology, ntf, latencies)
+    unit_module = UnitModule.from_full_ehi(ehi)
 
     instrs = [AssignCValueOp("x", 3)]
     subrt1 = simple_subroutine(
@@ -400,7 +322,29 @@ def test_initialize():
     """,
     )
 
-    program = create_program(instrs=instrs, subroutines={"subrt1": subrt1})
+    request_routine = RequestRoutine(
+        name="req1",
+        return_vars=[],
+        callback_type=CallbackType.WAIT_ALL,
+        callback=None,
+        request=QoalaRequest(
+            name="req1",
+            remote_id=1,
+            epr_socket_id=0,
+            num_pairs=1,
+            virt_ids=RequestVirtIdMapping.from_str("increment 0"),
+            timeout=1000,
+            fidelity=1.0,
+            typ=EprType.CREATE_KEEP,
+            role=EprRole.CREATE,
+        ),
+    )
+
+    program = create_program(
+        instrs=instrs,
+        subroutines={"subrt1": subrt1},
+        req_routines={"req1": request_routine},
+    )
     process = create_process(
         pid=0,
         program=program,
@@ -416,104 +360,51 @@ def test_initialize():
     assert host_mem.read("theta") == 3.14
     assert host_mem.read("name") == "alice"
 
-    request = NetstackCreateRequest(
-        remote_id=1,
-        epr_socket_id=0,
-        typ=EprCreateType.CREATE_KEEP,
-        num_pairs=1,
-        fidelity=1.0,
-        virt_qubit_ids=[0],
-        result_array_addr=0,
-    )
+    netsquid_run(host_processor.assign_instr_index(process, instr_idx=0))
+    routine = process.program.local_routines["subrt1"]
+    qnos_processor.instantiate_routine(process, routine, {}, 0, 0)
+    netsquid_run(qnos_processor.assign_routine_instr(process, "subrt1", 0))
 
-    netsquid_run(host_processor.assign(process, instr_idx=0))
-    netsquid_run(qnos_processor.assign(process, "subrt1", 0))
-
-    process.shared_mem.init_new_array(0, SER_RESPONSE_KEEP_LEN * 1)
-    netsquid_run(netstack_processor.assign(process, request))
+    rrcall = RrCallTuple.no_alloc("req1")
+    netsquid_run(netstack_processor.assign_request_routine(process, rrcall))
 
     assert process.host_mem.read("x") == 3
-    assert process.shared_mem.get_reg_value("R5") == 42
+    assert process.qnos_mem.get_reg_value("R5") == 42
     assert procnode.memmgr.phys_id_for(pid=process.pid, virt_id=0) == 0
 
 
 def test_2():
+    ns.sim_reset()
+
     num_qubits = 3
-    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})
+    topology = generic_topology(num_qubits)
+    latencies = LhiLatencies.all_zero()
+    ntf = GenericToVanillaInterface()
 
-    global_env = create_global_env(num_qubits)
-    procnode = create_procnode("alice", global_env, num_qubits)
-    procnode.qdevice = MockQDevice(topology)
-
-    # procnode.host.interface = MockHostInterface()
-
-    # mock_result = ResCreateAndKeep(bell_state=BellIndex.B01)
-    # procnode.netstack.interface = MockNetstackInterface(
-    #     local_env, procnode.qdevice, procnode.memmgr, mock_result
-    # )
-
-    host_processor = procnode.host.processor
-    qnos_processor = procnode.qnos.processor
-
-    unit_module = UnitModule.default_generic(num_qubits)
-
-    instrs = [RunSubroutineOp(None, IqoalaVector([]), "subrt1")]
-    subroutines = parse_iqoala_subroutines(
-        """
-SUBROUTINE subrt1
-    params: 
-    returns: R5 -> result
-    request:
-  NETQASM_START
-    set R5 42
-    ret_reg R5
-  NETQASM_END
-    """
+    network_info = create_network_info(num_qubits)
+    procnode = create_procnode(
+        "alice", network_info, num_qubits, topology, latencies, ntf, asynchronous=True
     )
-
-    program = create_program(instrs=instrs, subroutines=subroutines)
-    process = create_process(
-        pid=0,
-        program=program,
-        unit_module=unit_module,
-        host_interface=procnode.host._interface,
-        inputs={"x": 1, "theta": 3.14, "name": "alice"},
-    )
-    procnode.add_process(process)
-
-    host_processor.initialize(process)
-
-    netsquid_run(host_processor.assign(process, instr_idx=0))
-    netsquid_run(qnos_processor.assign(process, "subrt1", 0))
-    host_processor.copy_subroutine_results(process, "subrt1")
-
-    assert process.host_mem.read("result") == 42
-    assert process.shared_mem.get_reg_value("R5") == 42
-
-
-def test_2_async():
-    num_qubits = 3
-    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})
-
-    global_env = create_global_env(num_qubits)
-    procnode = create_procnode("alice", global_env, num_qubits, asynchronous=True)
     procnode.qdevice = MockQDevice(topology)
 
     host_processor = procnode.host.processor
     qnos_processor = procnode.qnos.processor
 
-    unit_module = UnitModule.default_generic(num_qubits)
+    ehi = LhiConverter.to_ehi(topology, ntf, latencies)
+    unit_module = UnitModule.from_full_ehi(ehi)
 
-    instrs = [RunSubroutineOp(None, IqoalaVector([]), "subrt1")]
+    instrs = [RunSubroutineOp(IqoalaVector(["result"]), IqoalaVector([]), "subrt1")]
     subroutines = parse_iqoala_subroutines(
         """
 SUBROUTINE subrt1
     params: 
-    returns: R5 -> result
+    returns: result
+    uses: 
+    keeps: 
     request:
   NETQASM_START
     set R5 42
-    ret_reg R5
+    store R5 @output[0]
   NETQASM_END
     """
     )
@@ -531,12 +422,10 @@ SUBROUTINE subrt1
     host_processor.initialize(process)
 
     def host_run() -> Generator[EventExpression, None, None]:
-        yield from host_processor.assign(process, instr_idx=0)
+        yield from host_processor.assign_instr_index(process, instr_idx=0)
 
     def qnos_run() -> Generator[EventExpression, None, None]:
-        yield from qnos_processor.assign(process, "subrt1", 0)
-        # Mock sending signal back to Host that subroutine has finished.
-        qnos_processor._interface.send_host_msg(Message(None))
+        yield from qnos_processor.await_local_routine_call(process)
 
     procnode.host.run = host_run
     procnode.qnos.run = qnos_run
@@ -544,30 +433,49 @@ SUBROUTINE subrt1
     ns.sim_run()
 
     assert process.host_mem.read("result") == 42
-    assert process.shared_mem.get_reg_value("R5") == 42
+    assert process.qnos_mem.get_reg_value("R5") == 42
 
 
 def test_classical_comm():
+    ns.sim_reset()
+
     num_qubits = 3
 
-    global_env = create_global_env(num_qubits)
+    topology = generic_topology(num_qubits)
+    latencies = LhiLatencies.all_zero()
+    ntf = GenericToVanillaInterface()
+
+    network_info = create_network_info(num_qubits)
 
     class TestProcNode(ProcNode):
         def run(self) -> Generator[EventExpression, None, None]:
             process = self.memmgr.get_process(0)
-            yield from self.host.processor.assign(process, 0)
+            yield from self.host.processor.assign_instr_index(process, 0)
 
     alice_procnode = create_procnode(
-        "alice", global_env, num_qubits, procnode_cls=TestProcNode
+        "alice",
+        network_info,
+        num_qubits,
+        topology,
+        latencies,
+        ntf,
+        procnode_cls=TestProcNode,
     )
     bob_procnode = create_procnode(
-        "bob", global_env, num_qubits, procnode_cls=TestProcNode
+        "bob",
+        network_info,
+        num_qubits,
+        topology,
+        latencies,
+        ntf,
+        procnode_cls=TestProcNode,
     )
 
     alice_host_processor = alice_procnode.host.processor
     bob_host_processor = bob_procnode.host.processor
 
-    unit_module = UnitModule.default_generic(num_qubits)
+    ehi = LhiConverter.to_ehi(topology, ntf, latencies)
+    unit_module = UnitModule.from_full_ehi(ehi)
 
     alice_instrs = [SendCMsgOp("csocket_id", "message")]
     alice_meta = ProgramMeta(
@@ -615,36 +523,61 @@ def test_classical_comm():
 
 
 def test_classical_comm_three_nodes():
+    ns.sim_reset()
+
     num_qubits = 3
 
-    global_env = create_global_env(num_qubits)
+    topology = generic_topology(num_qubits)
+    latencies = LhiLatencies.all_zero()
+    ntf = GenericToVanillaInterface()
+
+    network_info = create_network_info(num_qubits)
 
     class SenderProcNode(ProcNode):
         def run(self) -> Generator[EventExpression, None, None]:
             process = self.memmgr.get_process(0)
-            yield from self.host.processor.assign(process, 0)
+            yield from self.host.processor.assign_instr_index(process, 0)
 
     class ReceiverProcNode(ProcNode):
         def run(self) -> Generator[EventExpression, None, None]:
             process = self.memmgr.get_process(0)
-            yield from self.host.processor.assign(process, 0)
-            yield from self.host.processor.assign(process, 1)
+            yield from self.host.processor.assign_instr_index(process, 0)
+            yield from self.host.processor.assign_instr_index(process, 1)
 
     alice_procnode = create_procnode(
-        "alice", global_env, num_qubits, procnode_cls=SenderProcNode
+        "alice",
+        network_info,
+        num_qubits,
+        topology,
+        latencies,
+        ntf,
+        procnode_cls=SenderProcNode,
     )
     bob_procnode = create_procnode(
-        "bob", global_env, num_qubits, procnode_cls=SenderProcNode
+        "bob",
+        network_info,
+        num_qubits,
+        topology,
+        latencies,
+        ntf,
+        procnode_cls=SenderProcNode,
     )
     charlie_procnode = create_procnode(
-        "charlie", global_env, num_qubits, procnode_cls=ReceiverProcNode
+        "charlie",
+        network_info,
+        num_qubits,
+        topology,
+        latencies,
+        ntf,
+        procnode_cls=ReceiverProcNode,
     )
 
     alice_host_processor = alice_procnode.host.processor
     bob_host_processor = bob_procnode.host.processor
     charlie_host_processor = charlie_procnode.host.processor
 
-    unit_module = UnitModule.default_generic(num_qubits)
+    ehi = LhiConverter.to_ehi(topology, ntf, latencies)
+    unit_module = UnitModule.from_full_ehi(ehi)
 
     alice_instrs = [SendCMsgOp("csocket_id", "message")]
     alice_meta = ProgramMeta(
@@ -717,29 +650,84 @@ def test_classical_comm_three_nodes():
 
 
 def test_epr():
+    ns.sim_reset()
+
     num_qubits = 3
 
-    global_env = create_global_env(num_qubits)
-    alice_id = global_env.get_node_id("alice")
-    bob_id = global_env.get_node_id("bob")
+    topology = generic_topology(num_qubits)
+    latencies = LhiLatencies.all_zero()
+    ntf = GenericToVanillaInterface()
+
+    network_info = create_network_info(num_qubits)
+    alice_id = network_info.get_node_id("alice")
+    bob_id = network_info.get_node_id("bob")
 
     class TestProcNode(ProcNode):
         def run(self) -> Generator[EventExpression, None, None]:
             process = self.memmgr.get_process(0)
-            request = process.requests["req"]
-            yield from self.netstack.processor.assign(process, request)
+            rrcall = RrCallTuple.no_alloc("req1")
+            yield from self.netstack.processor.assign_request_routine(process, rrcall)
 
     alice_procnode = create_procnode(
-        "alice", global_env, num_qubits, procnode_cls=TestProcNode
+        "alice",
+        network_info,
+        num_qubits,
+        topology,
+        latencies,
+        ntf,
+        procnode_cls=TestProcNode,
     )
     bob_procnode = create_procnode(
-        "bob", global_env, num_qubits, procnode_cls=TestProcNode
+        "bob",
+        network_info,
+        num_qubits,
+        topology,
+        latencies,
+        ntf,
+        procnode_cls=TestProcNode,
     )
 
     alice_host_processor = alice_procnode.host.processor
     bob_host_processor = bob_procnode.host.processor
 
-    unit_module = UnitModule.default_generic(num_qubits)
+    ehi = LhiConverter.to_ehi(topology, ntf, latencies)
+    unit_module = UnitModule.from_full_ehi(ehi)
+
+    alice_request_routine = RequestRoutine(
+        name="req1",
+        return_vars=[],
+        callback_type=CallbackType.WAIT_ALL,
+        callback=None,
+        request=QoalaRequest(
+            name="req1",
+            remote_id=bob_id,
+            epr_socket_id=0,
+            num_pairs=1,
+            virt_ids=RequestVirtIdMapping.from_str("increment 0"),
+            timeout=1000,
+            fidelity=1.0,
+            typ=EprType.CREATE_KEEP,
+            role=EprRole.CREATE,
+        ),
+    )
+
+    bob_request_routine = RequestRoutine(
+        name="req1",
+        return_vars=[],
+        callback_type=CallbackType.WAIT_ALL,
+        callback=None,
+        request=QoalaRequest(
+            name="req1",
+            remote_id=alice_id,
+            epr_socket_id=0,
+            num_pairs=1,
+            virt_ids=RequestVirtIdMapping.from_str("increment 0"),
+            timeout=1000,
+            fidelity=1.0,
+            typ=EprType.CREATE_KEEP,
+            role=EprRole.RECEIVE,
+        ),
+    )
 
     alice_instrs = [SendCMsgOp("csocket_id", "message")]
     alice_meta = ProgramMeta(
@@ -748,7 +736,11 @@ def test_epr():
         csockets={0: "bob"},
         epr_sockets={},
     )
-    alice_program = create_program(instrs=alice_instrs, meta=alice_meta)
+    alice_program = create_program(
+        instrs=alice_instrs,
+        req_routines={"req1": alice_request_routine},
+        meta=alice_meta,
+    )
     alice_process = create_process(
         pid=0,
         program=alice_program,
@@ -763,7 +755,9 @@ def test_epr():
     bob_meta = ProgramMeta(
         name="bob", parameters=["csocket_id"], csockets={0: "alice"}, epr_sockets={}
     )
-    bob_program = create_program(instrs=bob_instrs, meta=bob_meta)
+    bob_program = create_program(
+        instrs=bob_instrs, req_routines={"req1": bob_request_routine}, meta=bob_meta
+    )
     bob_process = create_process(
         pid=0,
         program=bob_program,
@@ -776,40 +770,22 @@ def test_epr():
 
     alice_procnode.connect_to(bob_procnode)
 
-    alice_request = NetstackCreateRequest(
-        remote_id=bob_id,
-        epr_socket_id=0,
-        typ=EprCreateType.CREATE_KEEP,
-        num_pairs=1,
-        fidelity=1.0,
-        virt_qubit_ids=[0],
-        result_array_addr=0,
-    )
-
-    bob_request = NetstackReceiveRequest(
-        remote_id=alice_id,
-        epr_socket_id=0,
-        typ=EprCreateType.CREATE_KEEP,
-        num_pairs=1,
-        fidelity=1.0,
-        virt_qubit_ids=[0],
-        result_array_addr=0,
-    )
-
-    alice_egp, bob_egp = create_egp_protocols(alice_procnode.node, bob_procnode.node)
-    alice_procnode.egpmgr.add_egp(bob_id, alice_egp)
-    bob_procnode.egpmgr.add_egp(alice_id, bob_egp)
-    alice_process.requests = {"req": alice_request}
-    bob_process.requests = {"req": bob_request}
-
-    alice_process.shared_mem.init_new_array(0, 10)
-    bob_process.shared_mem.init_new_array(0, 10)
+    nodes = [alice_procnode.node, bob_procnode.node]
+    entdistcomp = EntDistComponent(network_info)
+    alice_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("alice"))
+    alice_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("alice"))
+    bob_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("bob"))
+    bob_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("bob"))
+    entdist = EntDist(nodes=nodes, network_info=network_info, comp=entdistcomp)
+    link_info = LhiLinkInfo.perfect(1000)
+    entdist.add_sampler(alice_procnode.node.ID, bob_procnode.node.ID, link_info)
 
     # First start Bob, since Alice won't yield on anything (she only does a Send
     # instruction) and therefore calling 'start()' on alice completes her whole
     # protocol while Bob's interface has not even been started.
     bob_procnode.start()
     alice_procnode.start()
+    entdist.start()
     ns.sim_run()
 
     assert alice_procnode.memmgr.phys_id_for(pid=0, virt_id=0) == 0
@@ -821,6 +797,8 @@ def test_epr():
 
 
 def test_whole_program():
+    ns.sim_reset()
+
     server_text = """
 META_START
     name: server
@@ -829,65 +807,51 @@ META_START
     epr_sockets: 0 -> client
 META_END
 
-remote_id = assign_cval() : {client_id}
-run_subroutine(vec<remote_id>) : subrt1
-
-SUBROUTINE subrt1
-    params: remote_id
-    returns:
-    request: req1
-  NETQASM_START
-    array 10 @0
-    recv_epr C15 C0 C1 C0
-    wait_all @0[0:10]
-  NETQASM_END
+^b0 {type = CL}:
+    remote_id = assign_cval() : {client_id}
+^b1 {type = QC}:
+    run_request(vec<>) : req1
 
 REQUEST req1
-  role: receive
+  callback_type: wait_all
+  callback:
+  return_vars: 
   remote_id: {client_id}
   epr_socket_id: 0
-  typ: create_keep
   num_pairs: 1
+  virt_ids: all 0
+  timeout:1000
   fidelity: 1.0
-  virt_qubit_ids: 0
-  result_array_addr: 0
+  typ: create_keep
+  role: receive
     """
-    server_program = IqoalaParser(server_text).parse()
+    server_program = QoalaParser(server_text).parse()
 
     num_qubits = 3
-    unit_module = UnitModule.default_generic(num_qubits)
-    global_env = create_global_env(num_qubits, names=["client", "server"])
-    server_id = global_env.get_node_id("server")
-    client_id = global_env.get_node_id("client")
+    topology = generic_topology(num_qubits)
+    latencies = LhiLatencies.all_zero()
+    ntf = GenericToVanillaInterface()
+    ehi = LhiConverter.to_ehi(topology, ntf, latencies)
+    unit_module = UnitModule.from_full_ehi(ehi)
+    network_info = create_network_info(num_qubits, names=["client", "server"])
+    server_id = network_info.get_node_id("server")
+    client_id = network_info.get_node_id("client")
 
     class TestProcNode(ProcNode):
         def run(self) -> Generator[EventExpression, None, None]:
             process = self.memmgr.get_process(0)
             self.scheduler.initialize_process(process)
-
-            subrt1 = process.subroutines["subrt1"]
-            epr_instr_idx = None
-            for i, instr in enumerate(subrt1.subroutine.instructions):
-                if isinstance(instr, CreateEPRInstruction) or isinstance(
-                    instr, RecvEPRInstruction
-                ):
-                    epr_instr_idx = i
-                    break
-
-            for i in range(epr_instr_idx):
-                yield from self.qnos.processor.assign(process, "subrt1", i)
-
-            request = process.requests["req1"].request
-            print("hello 1?")
-            yield from self.netstack.processor.assign(process, request)
-
-            # wait instr
-            yield from self.qnos.processor.assign(process, "subrt1", epr_instr_idx + 1)
-            print("hello 2?")
-            # yield from self.netstack.processor.assign(process, request)
+            rrcall = RrCallTuple.no_alloc("req1")
+            yield from self.netstack.processor.assign_request_routine(process, rrcall)
 
     server_procnode = create_procnode(
-        "server", global_env, num_qubits, procnode_cls=TestProcNode
+        "server",
+        network_info,
+        num_qubits,
+        topology,
+        latencies,
+        ntf,
+        procnode_cls=TestProcNode,
     )
     server_process = create_process(
         pid=0,
@@ -906,33 +870,33 @@ META_START
     epr_sockets: 0 -> server
 META_END
 
-remote_id = assign_cval() : {server_id}
-run_subroutine(vec<remote_id>) : subrt1
-
-SUBROUTINE subrt1
-    params: remote_id
-    returns:
-    request: req1
-  NETQASM_START
-    set C10 10
-    array C10 @0
-    create_epr C15 C0 C1 C2 C0
-    wait_all @0[0:10]
-  NETQASM_END
+^b0 {type = CL}:
+    remote_id = assign_cval() : {server_id}
+^b1 {type = QL}:
+    run_request(vec<>) : req1
 
 REQUEST req1
-  role: create
+  callback_type: wait_all
+  callback:
+  return_vars: 
   remote_id: {server_id}
   epr_socket_id: 0
-  typ: create_keep
   num_pairs: 1
+  virt_ids: all 0
+  timeout: 1000
   fidelity: 1.0
-  virt_qubit_ids: 0
-  result_array_addr: 0
+  typ: create_keep
+  role: create
     """
-    client_program = IqoalaParser(client_text).parse()
+    client_program = QoalaParser(client_text).parse()
     client_procnode = create_procnode(
-        "client", global_env, num_qubits, procnode_cls=TestProcNode
+        "client",
+        network_info,
+        num_qubits,
+        topology,
+        latencies,
+        ntf,
+        procnode_cls=TestProcNode,
     )
     client_process = create_process(
         pid=0,
@@ -943,17 +907,21 @@ REQUEST req1
     )
     client_procnode.add_process(client_process)
 
-    client_egp, server_egp = create_egp_protocols(
-        client_procnode.node, server_procnode.node
-    )
-    client_procnode.egpmgr.add_egp(server_id, client_egp)
-    server_procnode.egpmgr.add_egp(client_id, server_egp)
-
     client_procnode.connect_to(server_procnode)
 
-    # client_egp._ll_prot.start()
+    nodes = [client_procnode.node, server_procnode.node]
+    entdistcomp = EntDistComponent(network_info)
+    client_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("client"))
+    client_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("client"))
+    server_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("server"))
+    server_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("server"))
+    entdist = EntDist(nodes=nodes, network_info=network_info, comp=entdistcomp)
+    link_info = LhiLinkInfo.perfect(1000)
+    entdist.add_sampler(client_procnode.node.ID, server_procnode.node.ID, link_info)
+
     server_procnode.start()
     client_procnode.start()
+    entdist.start()
     ns.sim_run()
 
     assert client_procnode.memmgr.phys_id_for(pid=0, virt_id=0) == 0
@@ -969,7 +937,6 @@ REQUEST req1
 if __name__ == "__main__":
     test_initialize()
     test_2()
-    test_2_async()
     test_classical_comm()
     test_classical_comm_three_nodes()
     test_epr()
